/*
 * Copyright 2011-2019 Jiří Janoušek <janousek.jiri@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

namespace Nuvola.JSTools {
/**
 * Converts JavaScriptCore string to Vala UTF-8 string
 *
 * @param jsstring JavaScriptCore string
 * @return Vala UTF-8 string
 */
public static string utf8_string(JsCore.String jsstring) {
    size_t max_size = jsstring.get_maximum_utf8_string_size();
    uint8[] buffer = new uint8[max_size];
    jsstring.get_utf8_string(buffer);
    return (string) buffer;
}

/**
 * Creates JavaScript object from JSON.
 *
 * @param ctx   JavaScript context
 * @param json  object in JSON notation
 * @return  JavaScript object
 */
public unowned JsCore.Value object_from_JSON(JsCore.Context ctx, string json) {
    unowned JsCore.Value? result = JsCore.Value.from_JSON(ctx, new JsCore.String(json == "" ? "{}" : json));
    if (result == null || !result.is_object(ctx)) {
        result = ctx.make_object();
    }
    return result;
}

/**
 * Creates JavaScript Exception object.
 *
 * @param ctx   JavaScript context
 * @param message   Error message (single line);
 * @return  JavaScript object
 */
public unowned JsCore.Value create_exception(JsCore.Context ctx, string message) {
    string exception = """{"type":"NuvolaError", "message":"%s"}""".printf(message.replace("\"", "\\\""));
    debug(exception);
    return object_from_JSON(ctx, exception);
}

/**
 * Obtains string property of JavaScript object
 *
 * @param ctx JavaScript context
 * @param obj JavaScript object
 * @param property property name
 * @return string value or null if property is not a string
 */
public static string? o_get_string(JsCore.Context ctx, JsCore.Object obj, string property) {
    unowned JsCore.Value value = obj.get_property(ctx, new JsCore.String(property));
    if (value.is_string(ctx)) {
        return utf8_string(value.to_jsstring(ctx));
    }
    return null;
}

/**
 * Obtains number property of JavaScript object
 *
 * @param ctx JavaScript context
 * @param obj JavaScript object
 * @param property property name
 * @return double value or 0.0 if property is not a number
 */
public static double o_get_number(JsCore.Context ctx, JsCore.Object obj, string property) {
    unowned JsCore.Value value = obj.get_property(ctx, new JsCore.String(property));
    if (value.is_number(ctx)) {
        return value.to_number(ctx);
    }
    return 0.0;
}

/**
 * Sets string property of JavaScript object
 *
 * @param ctx JavaScript context
 * @param obj JavaScript object
 * @param property property name
 * @param value property value
 */
public static void o_set_string(JsCore.Context ctx, JsCore.Object obj, string property, string value) {
    obj.set_property(ctx, new JsCore.String(property), JsCore.Value.string(ctx, new JsCore.String(value)));
}

/**
 * Sets string property of JavaScript object
 *
 * @param ctx JavaScript context
 * @param obj JavaScript object
 * @param property property name
 * @param value property value
 */
public static void o_set_number(JsCore.Context ctx, JsCore.Object obj, string property, double value) {
    obj.set_property(ctx, new JsCore.String(property), JsCore.Value.number(ctx, value));
}

/**
 * Sets null property of JavaScript object
 *
 * @param ctx JavaScript context
 * @param obj JavaScript object
 * @param property property name
 */
public static void o_set_null(JsCore.Context ctx, JsCore.Object obj, string property) {
    obj.set_property(ctx, new JsCore.String(property), JsCore.Value.null(ctx));
}

/**
 * Sets boolean property of JavaScript object
 *
 * @param ctx JavaScript context
 * @param obj JavaScript object
 * @param property property name
 * @param value property value
 */
public static void o_set_bool(JsCore.Context ctx, JsCore.Object obj, string property, bool value) {
    obj.set_property(ctx, new JsCore.String(property), JsCore.Value.boolean(ctx, value));
}

/**
 * Obtains object property of JavaScript object
 *
 * @param ctx JavaScript context
 * @param obj JavaScript object
 * @param property property name
 * @return object or null if property is not a object
 */
public static unowned JsCore.Object? o_get_object(JsCore.Context ctx, JsCore.Object obj, string property) {
    unowned JsCore.Value value = obj.get_property(ctx, new JsCore.String(property));
    if (value.is_object(ctx)) {
        return value.to_object(ctx);
    }
    return null;
}

/**
 * Converts JavaScript value to string
 *
 * @param ctx            JavaScript context
 * @param val            JavaScript value
 * @param allow_empty    if false null will be returned for empty string
 * @return string value or null if value is not a string
 */
public static string? string_or_null(JsCore.Context ctx, JsCore.Value val, bool allow_empty=false) {
    if (val.is_string(ctx)) {
        string str = utf8_string(val.to_jsstring(ctx));
        return (str == "" && !allow_empty) ? null : str;
    }
    return null;
}

public string? value_to_string(JsCore.Context ctx, JsCore.Value value) {
    if (value.is_string(ctx)) {
        return utf8_string(value.to_jsstring(ctx));
    }
    if (value.is_number(ctx)) {
        return (value.to_number(ctx).to_string());
    }
    if (value.is_object(ctx)) {
        return utf8_string(value.to_object(ctx).to_JSON(ctx, 0, null));
    }
    return null;
}

public string? exception_to_string(JsCore.Context ctx, JsCore.Value value) {
    if (value.is_object(ctx)) {
        unowned JsCore.Object obj = value.to_object(ctx);
        string? message = o_get_string(ctx, obj, "message");
        if (message != null) {
            string? name = o_get_string(ctx, obj, "name");
            var line = (int) o_get_number(ctx, obj, "line");
            string? file = o_get_string(ctx, obj, "sourceURL");
            if (line == 0 && file == null) {
                return "%s: %s. Enable JS debugging for more details.".printf(name ?? "null", message);
            }
            return "%s:%d: %s: %s".printf(file ?? "(null)", line, name ?? "null", message);
        }
    }
    return value_to_string(ctx, value);
}

public unowned JsCore.Value get_gobject_property_named(JsCore.Context ctx, GLib.Object o, string name) {
    ObjectClass klass = (ObjectClass) o.get_type().class_ref();
    unowned ParamSpec? p = klass.find_property(name);
    if (p == null) {
        return JsCore.Value.undefined(ctx);
    }
    return get_gobject_property(ctx, o, p);
}

public unowned JsCore.Value get_gobject_property(JsCore.Context ctx, GLib.Object o, ParamSpec p) {
    GLib.Type type = p.value_type;
    if (type == typeof(string)) {
        string str_val;
        o.@get(p.name, out str_val);
        return JsCore.Value.string(ctx, new JsCore.String(str_val));
    }

    if (type == typeof(int)) {
        int int_val;
        o.@get(p.name, out int_val);
        return JsCore.Value.number(ctx, (double) int_val);
    }

    if (type == typeof(float)) {
        float float_val;
        o.@get(p.name, out float_val);
        return JsCore.Value.number(ctx, (double) float_val);
    }

    if (type == typeof(double)) {
        double double_val;
        o.@get(p.name, out double_val);
        return JsCore.Value.number(ctx, (double) double_val);
    }

    if (type == typeof(bool)) {
        bool bool_val;
        o.@get(p.name, out bool_val);
        return JsCore.Value.boolean(ctx, bool_val);
    }

    return JsCore.Value.undefined(ctx);
}

public unowned JsCore.Value value_from_variant(JsCore.Context ctx, Variant? variant) throws JSError {
    if (variant == null) {
        return JsCore.Value.null(ctx);
    }

    VariantType type = variant.get_type();

    if (variant.is_of_type(VariantType.VARIANT)) {
        return value_from_variant(ctx, variant.get_variant());
    }

    if (type.is_subtype_of(VariantType.MAYBE)) {
        Variant? maybe_variant = null;
        variant.get("m*", &maybe_variant);
        if (maybe_variant == null) {
            return JsCore.Value.null(ctx);
        }
        return value_from_variant(ctx, maybe_variant);
    }

    var object_type = new VariantType("a{s*}");
    if (type.is_subtype_of(object_type)) {
        unowned JsCore.Object object = ctx.make_object();
        VariantIter iter = null; // "a{s*}" (new allocation)
        variant.get("a{s*}", out iter);
        unowned string? key = null; // "&s" (unowned)
        Variant? value = null; // "*" (new reference)
        while (iter.next("{&s*}", out key, out value)) {
            object.set_property(ctx, new JsCore.String(key), value_from_variant(ctx, value));
            value = null; // https://gitlab.gnome.org/GNOME/vala/issues/722
        }
        return object;
    }

    if (variant.is_of_type(VariantType.STRING)) {
        return JsCore.Value.string(ctx, new JsCore.String(variant.get_string()));
    }

    if (variant.is_of_type(VariantType.BOOLEAN)) {
        return JsCore.Value.boolean(ctx, variant.get_boolean());
    }

    if (variant.is_of_type(VariantType.DOUBLE)) {
        return JsCore.Value.number(ctx, variant.get_double());
    }

    if (variant.is_of_type(VariantType.INT32)) {
        return JsCore.Value.number(ctx, (double) variant.get_int32());
    }

    if (variant.is_of_type(VariantType.UINT32)) {
        return JsCore.Value.number(ctx, (double) variant.get_uint32());
    }

    if (variant.is_of_type(VariantType.INT64)) {
        return JsCore.Value.number(ctx, (double) variant.get_int64());
    }

    if (variant.is_of_type(VariantType.UINT64)) {
        return JsCore.Value.number(ctx, (double) variant.get_uint64());
    }

    if (variant.is_container()) {
        size_t size = variant.n_children();
        void*[] args = new void*[size];
        for (var i = 0; i < size; i++) {
            args[i] = (void*) value_from_variant(ctx, variant.get_child_value(i));
        }
        return ctx.make_array((JsCore.Value[]) args);
    }

    throw new JSError.WRONG_TYPE("Unsupported type '%s'. Content: %s", variant.get_type_string(), variant.print(true));
}

public Variant variant_from_value(JsCore.Context ctx, JsCore.Value val) throws JSError {
    if (val.is_null(ctx)) {
        return new Variant("mv", null);
    }

    if (val.is_string(ctx)) {
        return new Variant.string(utf8_string(val.to_jsstring(ctx)));
    }

    if (val.is_number(ctx)) {
        return new Variant.double(val.to_number(ctx));
    }

    if (val.is_boolean(ctx)) {
        return new Variant.boolean(val.to_boolean(ctx));
    }

    unowned JsCore.Object glob_object = ctx.get_global_object();
    unowned JsCore.Object object = o_get_object(ctx, glob_object, "Array");
    object = o_get_object(ctx, object, "isArray");
    void*[] params = new void*[1];
    params[0] = (void*) val;
    JsCore.Value? exception;
    unowned JsCore.Value result = object.call_as_function(ctx, glob_object, (JsCore.Value[]) params, out exception);
    if (exception != null) {
        throw new JSError.WRONG_TYPE("Unsupported type. %s", exception_to_string(ctx, exception) ?? "(null)");
    }
    if (result.to_boolean(ctx)) {
        VariantBuilder builder = new VariantBuilder(new VariantType ("av"));
        object = (JsCore.Object) val;
        int size = (int) o_get_number(ctx, object, "length");
        for (uint i = 0; i < size; i++) {
            builder.add("v", variant_from_value(ctx, object.get_property_at_index(ctx, i)));
        }

        return builder.end();
    }

    if (val.is_object(ctx)) {
        object = (JsCore.Object) val;
        JsCore.Properties properties = object.get_properties(ctx);
        size_t size = properties.get_count();
        var builder = new VariantBuilder(new VariantType("a{smv}"));
        for (size_t i = 0; i < size; i++) {
            unowned JsCore.String js_property = properties.get(i);
            Variant? value = variant_from_value(ctx, object.get_property(ctx, js_property));
            builder.add("{smv}", utf8_string(js_property), value);
        }

        return builder.end();
    }
    warning("Attempt to convert `undefined` JavaScript type. This might be a programmer error!");
    return new Variant.string("<ERROR: UNDEFINED VALUE TYPE>");
}

} // namespace Nuvola.JSTools
